// Copyright (c) 2014 Oyster
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

package halfshell

import (
	"fmt"
	"io/ioutil"
	"mime"
	"net/http"
	"os"
	"path/filepath"
	"strconv"
	"strings"
)

// Image contains a byte array of the image data and its MIME type.
// TODO: See if we can use the std library's Image type without incurring
// the hit of extra copying.
type Image struct {
	Bytes     []byte
	MimeType  string
	Signature string
}

// NewImageFromHTTPResponse returns a pointer to a new Image created from an
// HTTP response object.
func NewImageFromHTTPResponse(httpResponse *http.Response) (*Image, error) {
	imageBytes, err := ioutil.ReadAll(httpResponse.Body)
	defer httpResponse.Body.Close()
	if err != nil {
		return nil, err
	}

	return &Image{
		Bytes:    imageBytes,
		MimeType: httpResponse.Header.Get("Content-Type"),
	}, nil
}

// NewImageFromFile returns a pointer to a new Image created from a file.
func NewImageFromFile(file *os.File) (*Image, error) {
	imageBytes, err := ioutil.ReadAll(file)
	if err != nil {
		return nil, err
	}

	return &Image{
		Bytes:    imageBytes,
		MimeType: mime.TypeByExtension(filepath.Ext(file.Name())),
	}, nil
}

// ImageDimensions is the width and height of an image.
type ImageDimensions struct {
	Width  uint64
	Height uint64
}

// Focalpoint is an x/y pair representing the location of the image subject.
// 0.5/0.5 is the middle.
type Focalpoint struct {
	X float64
	Y float64
}

// NewFocalpoint splits the given string into a Focalpoint struct. The string
// format should be: "X,Y". For example: "0.1,0.1".
func NewFocalpoint(s string) (fp Focalpoint) {
	fp = Focalpoint{0.5, 0.5}

	pair := strings.Split(s, ",")
	if len(pair) != 2 {
		return
	}

	x, _ := strconv.ParseFloat(pair[0], 64)
	y, _ := strconv.ParseFloat(pair[1], 64)

	fp.X = x
	fp.Y = y
	return
}

// AspectRatio returns the image dimension's aspect ratio.
func (d ImageDimensions) AspectRatio() float64 {
	return float64(d.Width) / float64(d.Height)
}

func (d ImageDimensions) String() string {
	return fmt.Sprintf("%dx%d", d.Width, d.Height)
}
